@@ -26,7 +26,7 @@ import unicodedata
 
 from com.dtmilano.android.adb.dumpsys import Dumpsys
 
-__version__ = '20.7.1'
+__version__ = '21.4.3'
 
 import sys
 import warnings
@@ -921,6 +921,22 @@ class AdbClient:
             profileEnd()
         return image
 
+    def imageToData(self, image, output_type=None):
+        """
+        Helps in cases where the Views cannot be identified.
+        Returns the text found in the image and its bounding box.
+        :param image: the image (i.e. from takeScreenshot())
+        :param output_type: the output type (defualt: pytessearct.Output.DICT)
+        :return: the data from the image
+        """
+        try:
+            import pytessearct
+        except ImportError:
+            raise Exception("You have to install pytesseract to use imageToData()")
+        if not output_type:
+            output_type = pytessearct.Output.DICT
+        return pytessearct.image_to_data(image, output_type=output_type)
+
     def __transformPointByOrientation(self, xxx_todo_changeme, orientationOrig, orientationDest):
         (x, y) = xxx_todo_changeme
         if orientationOrig != orientationDest:
