@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 '''
-Copyright (C) 2012-2019  Diego Torres Milano
+Copyright (C) 2012-2022  Diego Torres Milano
 Created on oct 6, 2014
 
 Licensed under the Apache License, Version 2.0 (the "License");
@@ -31,9 +31,10 @@ from culebratester_client import WindowHierarchy
 from com.dtmilano.android.common import profileEnd
 from com.dtmilano.android.common import profileStart
 from com.dtmilano.android.concertina import Concertina
-from com.dtmilano.android.viewclient import ViewClient
+from com.dtmilano.android.keyevent import KEY_EVENT
+from com.dtmilano.android.viewclient import ViewClient, View
 
-__version__ = '20.7.1'
+__version__ = '21.4.3'
 
 import sys
 import threading
@@ -130,6 +131,7 @@ class Operation:
     OPEN_QUICK_SETTINGS = 'open_quick_settings'
     TYPE = 'type'
     PRESS = 'press'
+    PRESS_UI_AUTOMATOR_HELPER = 'press_ui_automator_helper'
     PRESS_BACK = 'press_back'
     PRESS_BACK_UI_AUTOMATOR_HELPER = 'press_back_ui_automator_helper'
     PRESS_HOME = 'press_home'
@@ -140,10 +142,13 @@ class Operation:
     SET_TEXT = 'set_text'
     SNAPSHOT = 'snapshot'
     START_ACTIVITY = 'start_activity'
+    START_ACTIVITY_UI_AUTOMATOR_HELPER = 'start_activity_ui_automator__helper'
     SLEEP = 'sleep'
     SWIPE_UI_AUTOMATOR_HELPER = 'swipe_ui_automator_helper'
     TRAVERSE = 'traverse'
     VIEW_SNAPSHOT = 'view_snapshot'
+    WAIT_FOR_IDLE_UI_AUTOMATOR_HELPER = 'wait_for_idle_ui_automator_helper'
+    WAIT_FOR_WINDOW_UPDATE_UI_AUTOMATOR_HELPER = 'wait_for_window_update_ui_automator_helper'
     WAKE = 'wake'
 
     COMMAND_NAME_OPERATION_MAP = {'flingBackward': FLING_BACKWARD, 'flingForward': FLING_FORWARD,
@@ -167,6 +172,7 @@ class Culebron:
 
     KEYSYM_TO_KEYCODE_MAP = {
         'Home': 'HOME',
+        'abovedot': 'HOME',  # Option+H on macOS
         'BackSpace': 'BACK',
         'Left': 'DPAD_LEFT',
         'Right': 'DPAD_RIGHT',
@@ -252,6 +258,7 @@ This is usually installed by python package. Check your distribution details.
         '''
 
         self.vc = vc
+        self.dump = None
         if 'CONCERTINA' in self.vc.debug:
             global DEBUG_CONCERTINA
             DEBUG_CONCERTINA = self.vc.debug['CONCERTINA'] is not None
@@ -336,12 +343,14 @@ This is usually installed by python package. Check your distribution details.
         if DEBUG:
             print("takeScreenshotAndShowItOnWindow()", file=sys.stderr)
         if self.vc and self.vc.uiAutomatorHelper:
-            received = self.vc.uiAutomatorHelper.takeScreenshot()
-            if sys.version_info[0] < 3:
-                stream = io.StringIO(received)
-            else:
-                stream = io.BytesIO(received.data)
-            self.unscaledScreenshot = Image.open(stream)
+            received = self.vc.uiAutomatorHelper.ui_device.take_screenshot()
+            stream = io.BytesIO(received.read())
+            try:
+                self.unscaledScreenshot = Image.open(stream)
+            except IOError as ex:
+                print(ex, file=sys.stderr)
+                print(repr(stream))
+                sys.exit(1)
         else:
             self.unscaledScreenshot = self.device.takeSnapshot(reconnect=True)
         self.image = self.unscaledScreenshot
@@ -603,15 +612,15 @@ This is usually installed by python package. Check your distribution details.
                 parent_unique_id = parent.getUniqueId()
         except AttributeError:
             vuid = view.unique_id
-            #FIXME
-            #is_target = view.is_target
+            # FIXME
+            # is_target = view.is_target
             is_target = False
             text = view.text
             parent = view.parent
             if parent:
-                #FIXME:
+                # FIXME:
                 # parent is an int
-                #parent_unique_id = parent.unique_id
+                # parent_unique_id = parent.unique_id
                 parent_unique_id = 0
         if parent is None:
             self.viewTree.insert('', tkinter.END, vuid, text=text)
@@ -644,7 +653,8 @@ This is usually installed by python package. Check your distribution details.
             window = -1
         if self.vc:
             dump = self.vc.dump(window=window, sleep=0.1)
-            self.printOperation(None, Operation.DUMP, window, dump)
+            if not self.vc.uiAutomatorHelper:
+                self.printOperation(None, Operation.DUMP, window, dump)
         else:
             dump = []
         self.dump = dump
@@ -674,7 +684,7 @@ This is usually installed by python package. Check your distribution details.
 
         # FIXME: we are not populating the view tree now
         # there are some problems with culebratester2
-        #if self.vc:
+        # if self.vc:
         #    self.vc.traverse(transform=self.populateViewTree)
 
     def getViewContainingPointAndGenerateTestCondition(self, x, y):
@@ -785,7 +795,10 @@ This is usually installed by python package. Check your distribution details.
                 candidate = candidates[0]
                 self.touchView(candidate, v if candidate != v else None)
 
-        self.printOperation(None, Operation.SLEEP, Operation.DEFAULT)
+        if self.vc.uiAutomatorHelper:
+            self.printOperation(None, Operation.WAIT_FOR_WINDOW_UPDATE_UI_AUTOMATOR_HELPER, Operation.DEFAULT)
+        else:
+            self.printOperation(None, Operation.SLEEP, Operation.DEFAULT)
         self.vc.sleep(5)
         self.takeScreenshotAndShowItOnWindow()
 
@@ -833,10 +846,12 @@ This is usually installed by python package. Check your distribution details.
         ViewClient.sayText(text)
         self.printOperation(None, Operation.SAY_TEXT, text)
 
-    def touchView(self, v, root=None):
+    def touchView(self, v: View, root=None) -> None:
+        # FIXME: v.touch() handles the 2 cases for CulebraTester2-public and adbclient
+        # also, we should obtain the selector only once
         v.touch()
         if v.uiAutomatorHelper:
-            self.printOperation(v, Operation.TOUCH_VIEW_UI_AUTOMATOR_HELPER, v.obtainSelectorForView())
+            self.printOperation(v, Operation.TOUCH_VIEW_UI_AUTOMATOR_HELPER, v.obtain_selector())
         else:
             # we pass root=v as an argument so the corresponding findView*() searches in this
             # subtree instead of the full tree
@@ -908,10 +923,12 @@ This is usually installed by python package. Check your distribution details.
             self.statusBar.clear()
             return
 
-    def longTouchView(self, v, root=None):
+    def longTouchView(self, v: View, root=None) -> None:
+        # FIXME: v.longTouch() handles the 2 cases for CulebraTester2-public and adbclient
+        # also, we should obtain the selector only once
         v.longTouch()
         if v.uiAutomatorHelper:
-            self.printOperation(v, Operation.LONG_TOUCH_VIEW_UI_AUTOMATOR_HELPER, v.obtainSelectorForView())
+            self.printOperation(v, Operation.LONG_TOUCH_VIEW_UI_AUTOMATOR_HELPER, v.obtain_selector())
         else:
             # we pass root=v as an argument so the corresponding findView*() searches in this
             # subtree instead of the full tree
@@ -949,7 +966,7 @@ This is usually installed by python package. Check your distribution details.
 
     def onCtrlButton1Pressed(self, event):
         if DEBUG:
-            print("onCtrlButton1Pressed((", event.x, ", ", event.y, "))", file=sys.stderr)
+            print(f"onCtrlButton1Pressed(({event.x}, {event.y}))", file=sys.stderr)
         (scaledX, scaledY) = (event.x / self.scale, event.y / self.scale)
         l = self.vc.findViewsContainingPoint((scaledX, scaledY))
         if l and len(l) > 0:
@@ -971,47 +988,65 @@ This is usually installed by python package. Check your distribution details.
             print("onButton3Pressed((", event.x, ", ", event.y, "))", file=sys.stderr)
         self.showPopupMenu(event)
 
-    def command(self, keycode):
-        '''
+    def command(self, keycode: str) -> None:
+        """
         Presses a key.
         Generates the actual key press on the device and prints the line in the script.
-        '''
 
-        self.device.press(keycode)
-        self.printOperation(None, Operation.PRESS, keycode)
+        :param keycode the keycode name
+        """
+
+        if self.vc.uiAutomatorHelper:
+            try:
+                if not keycode.startswith('KEYCODE_'):
+                    keycode = f'KEYCODE_{keycode}'
+                self.vc.uiAutomatorHelper.ui_device.press_key_code(KEY_EVENT[f'{keycode}'])
+                self.printOperation(None, Operation.PRESS_UI_AUTOMATOR_HELPER, keycode)
+            except Exception as e:
+                print(e, file=sys.stderr)
+        else:
+            self.device.press(keycode)
+            self.printOperation(None, Operation.PRESS, keycode)
 
     def onKeyPressed(self, event):
         if DEBUG_KEY:
             print("onKeyPressed(", repr(event), ")", file=sys.stderr)
-            print("    event", type(event.char), len(event.char), repr(
-                event.char), "keysym=", event.keysym, "keycode=", event.keycode, event.type, "state=", event.state,
+            print(f'    char: type={type(event.char)} len={len(event.char)} repr={repr(event.char)}\n' +
+                  f'    keysym={event.keysym}, keycode={event.keycode}, type={event.type}, state={event.state}',
                   file=sys.stderr)
             print("    events disabled:", self.areEventsDisabled, file=sys.stderr)
         if self.areEventsDisabled:
             if DEBUG_KEY:
-                print("ignoring event", file=sys.stderr)
+                print("    ignoring event", file=sys.stderr)
             self.canvas.update_idletasks()
             return
 
         char = event.char
         keysym = event.keysym
+        state = event.state
+
+        # Manual way to get the modifiers
+        # https://stackoverflow.com/a/34482048/236465
+        ctrl = (state & 0x4) != 0
+        alt = (state & 0x8) != 0 or (state & 0x80) != 0
+        shift = (state & 0x1) != 0
+
+        if ctrl:
+            try:
+                return getattr(self, f'onCtrl{keysym.upper()}')(event)
+            except AttributeError:
+                pass
 
         if len(char) == 0 and not (
                 keysym in Culebron.KEYSYM_TO_KEYCODE_MAP or keysym in Culebron.KEYSYM_CULEBRON_COMMANDS):
             if DEBUG_KEY:
-                print("returning because len(char) == 0", file=sys.stderr)
+                print(f'    returning because len(char) == 0 keysym={keysym}', file=sys.stderr)
             return
 
         ###
         ### internal commands: no output to generated script
         ###
-        try:
-            handler = getattr(self, 'onCtrl%s' % self.UPPERCASE_CHARS[ord(char) - 1])
-        except:
-            handler = None
-        if handler:
-            return handler(event)
-        elif keysym == 'F1':
+        if keysym == 'F1':
             self.showHelp()
             return
         elif keysym == 'F5':
@@ -1052,14 +1087,14 @@ This is usually installed by python package. Check your distribution details.
             else:
                 self.command(Culebron.KEYSYM_TO_KEYCODE_MAP[keysym])
         # ALT-M
-        elif keysym == 'm' and event.state == 24:
+        elif keysym == 'm' and alt:
             if DEBUG_KEY:
                 print("Sending MENU", file=sys.stderr)
             self.command('MENU')
         # OPTION-ENTER (mac)
         elif keysym == 'Return' and event.state == 16:
             if DEBUG_KEY:
-                print >> sys.stderr, "Sending DPAD_CENTER"
+                print("Sending DPAD_CENTER", file=sys.stderr)
             self.command('DPAD_CENTER')
         elif char == '\r':
             self.command('ENTER')
@@ -1067,9 +1102,7 @@ This is usually installed by python package. Check your distribution details.
             # do nothing
             pass
         else:
-            self.command(char.decode('ascii', errors='replace'))
-        # commented out (profile)
-        # time.sleep(1)
+            self.command(char)
         self.takeScreenshotAndShowItOnWindow()
 
     def wake(self):
@@ -1105,7 +1138,10 @@ This is usually installed by python package. Check your distribution details.
             self.toggleGenerateTestCondition()
 
     def printStartActivityAtTop(self):
-        self.printOperation(None, Operation.START_ACTIVITY, self.device.getTopActivityName())
+        if self.vc.uiAutomatorHelper:
+            self.printOperation(None, Operation.START_ACTIVITY_UI_AUTOMATOR_HELPER, self.device.getTopActivityName())
+        else:
+            self.printOperation(None, Operation.START_ACTIVITY, self.device.getTopActivityName())
 
     def onCtrlA(self, event):
         if DEBUG:
@@ -1148,7 +1184,8 @@ This is usually installed by python package. Check your distribution details.
 
         if not view:
             raise ValueError("view must be provided to take snapshot")
-        filename = self.snapshotDir + os.sep + '${serialno}-' + view.variableNameFromId() + '-${timestamp}' + '.' + self.snapshotFormat.lower()
+        filename = self.snapshotDir + os.sep + '${serialno}-' + View.variableNameFromId(
+            view) + '-${timestamp}' + '.' + self.snapshotFormat.lower()
         d = FileDialog(self, self.device.substituteDeviceTemplate(filename))
         saveAsFilename = d.askSaveAsFilename()
         if saveAsFilename:
@@ -1951,6 +1988,8 @@ if TKINTER_AVAILABLE:
             self.transient(self.parent)
             self.culebron.setDragDialogShowed(True)
             self.title("Drag: selecting parameters")
+            self.__grabbing = None
+            self.ok = None
 
             # valid percent substitutions (from the Tk entry man page)
             # %d = Type of action (1=insert, 0=delete, -1 for others)
@@ -2300,7 +2339,7 @@ if TKINTER_AVAILABLE:
             self.bind("<Return>", self.onDismiss)
             self.bind("<Escape>", self.onDismiss)
 
-            box.grid(row=1, column=1)
+            box.grid(row=2, column=1)
 
         def onDismiss(self, event=None):
             # put focus back to the parent window's canvas
