@@ -20,7 +20,7 @@ limitations under the License.
 
 from __future__ import print_function
 
-__version__ = '20.7.1'
+__version__ = '21.4.3'
 
 import json
 import os
@@ -93,16 +93,26 @@ class UiAutomatorHelper:
     TEST_CLASS = PACKAGE + '.test'
     TEST_RUNNER = 'com.dtmilano.android.uiautomatorhelper.UiAutomatorHelperTestRunner'
 
-    def __init__(self, adbclient, adb=None, localport=9987, remoteport=9987, hostname='localhost'):
+    def __init__(self, adbclient, adb=None, localport=9987, remoteport=9987, hostname='localhost', api_version='v2'):
+        """
+        UiAutomatorHelper constructor used when the backend selected is **CulebraTester2-public**.
+        This class holds references to the different API's:
+         - Device
+         - TargetContext
+         - ObjectStore
+         - UiDevice
+         - UiObject2
+         - Until
+
+        :param adbclient: the adb client
+        :param adb: adb if known
+        :param localport: the local port used in CulebraTester2-public port forwarding
+        :param remoteport: the remote port used in CulebraTester2-public port forwarding
+        :param hostname: the hostname used by CulebraTester2-public
+        :param api_version: the api version
+        """
         self.adbClient = adbclient
         ''' The adb client (a.k.a. device) '''
-        # instrumentation = self.adbClient.shell('pm list instrumentation %s' % self.PACKAGE)
-        # if not instrumentation:
-        #    raise RuntimeError('The target device does not contain the instrumentation for %s' % self.PACKAGE)
-        # if not re.match('instrumentation:%s/%s \(target=%s\)' % (self.TEST_CLASS, self.TEST_RUNNER, self.PACKAGE),
-        #                instrumentation):
-        #    raise RuntimeError('The instrumentation found for %s does not match the expected %s/%s' % (
-        #    self.PACKAGE, self.TEST_CLASS, self.TEST_RUNNER))
         self.adb = self.__whichAdb(adb)
         ''' The adb command '''
         self.osName = platform.system()
@@ -111,18 +121,13 @@ class UiAutomatorHelper:
         ''' Is it Mac OSX? '''
         self.hostname = hostname
         ''' The hostname we are connecting to. '''
-        # if hostname in ['localhost', '127.0.0.1']:
-        #    self.__redirectPort(localport, remoteport)
-        # self.__runTests()
-        # self.baseUrl = 'http://%s:%d' % (hostname, localport)
-        # try:
-        #    self.session = self.__connectSession()
-        # except RuntimeError as ex:
-        #    self.thread.forceStop()
-        #    raise ex
-        print('⚠️ CulebraTester2 server should have been started and port redirected.', file=sys.stderr)
-        # TODO: localport should be in ApiClient configuration
-        self.api_instance = culebratester_client.DefaultApi(culebratester_client.ApiClient())
+
+        print(
+            f'⚠️  CulebraTester2 server should have been started and localport {localport} redirected to remote port {remoteport}.',
+            file=sys.stderr)
+        configuration = culebratester_client.Configuration()
+        configuration.host = f'http://{hostname}:{localport}/{api_version}'
+        self.api_instance = culebratester_client.DefaultApi(culebratester_client.ApiClient(configuration))
         self.device: UiAutomatorHelper.Device = UiAutomatorHelper.Device(self)
         self.target_context: UiAutomatorHelper.TargetContext = UiAutomatorHelper.TargetContext(self)
         self.object_store: UiAutomatorHelper.ObjectStore = UiAutomatorHelper.ObjectStore(self)
@@ -192,7 +197,7 @@ class UiAutomatorHelper:
             print("__runTests: end", file=sys.stderr)
 
     def __httpCommand(self, url, params=None, method='GET'):
-        raise RuntimeError("this method should not be used")
+        raise RuntimeError(f"this method should not be used: url={url} params={params} method={method}")
 
     #
     # UiAutomatorHelper internal commands
@@ -268,16 +273,17 @@ class UiAutomatorHelper:
         def __init__(self, uiAutomatorHelper) -> None:
             super().__init__(uiAutomatorHelper)
 
-        def start_activity(self, pkg, cls):
+        def start_activity(self, pkg, cls, **kwargs):
             """
             Starts an activity.
 
             :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
             :param pkg: the package
             :param cls: the Activity class
+            :param kwargs: uri: the optional URI
             :return: the api response
             """
-            return self.uiAutomatorHelper.api_instance.target_context_start_activity_get(pkg, cls)
+            return self.uiAutomatorHelper.api_instance.target_context_start_activity_get(pkg, cls, **kwargs)
 
     #
     # ObjectStore
@@ -358,6 +364,8 @@ class UiAutomatorHelper:
             :param kwargs:
             :return:
             """
+            if 'body' in kwargs:
+                return self.uiAutomatorHelper.api_instance.ui_device_find_object_post(**kwargs)
             if self.some(['resource_id', 'ui_selector', 'by_selector'], kwargs):
                 return self.uiAutomatorHelper.api_instance.ui_device_find_object_get(**kwargs)
             body = culebratester_client.Selector(**kwargs)
@@ -382,6 +390,15 @@ class UiAutomatorHelper:
             """
             return self.uiAutomatorHelper.api_instance.ui_device_press_back_get()
 
+        def press_enter(self):
+            """
+            Presses ENTER.
+
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :return:
+            """
+            return self.uiAutomatorHelper.api_instance.ui_device_press_enter_get()
+
         def press_home(self):
             """
             Presses HOME.
@@ -449,6 +466,17 @@ class UiAutomatorHelper:
             """
             return self.uiAutomatorHelper.api_instance.ui_device_wait_for_idle_get(**kwargs)
 
+        def wait_for_window_update(self, timeout=5000, **kwargs):
+            """
+            Waits for window update.
+
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :param timeout: the timeout
+            :param kwargs:
+            :return:
+            """
+            return self.uiAutomatorHelper.api_instance.ui_device_wait_for_window_update_get(timeout, **kwargs)
+
     #
     # UiObject2
     #
@@ -517,7 +545,7 @@ class UiAutomatorHelper:
             :param oid: the oid
             :return: the result of the operation
             """
-            return self.uiAutomatorHelper.api_instance.ui_object2_long_click_get(oid=oid)
+            return self.uiAutomatorHelper.api_instance.ui_object2_oid_long_click_get(oid=oid)
 
         def set_text(self, oid, text):
             """
